' Program: COMPRESS.BAS
' Author: Gary Conway
' Modified by: Blake Ragsdell and Douglas H. Quebbeman
'
' This program will perform run-length encoding 
' on an input file.
'
' Next, it decompresses a file that has be RLE encoded. Compressed
' byte sequences are decoded into their original form. In this
' example, the names are hard-coded into the program.
'
' During compression, the input filename is COMPRESS.BAS and the
' ouput filename is OUTPUT. For decompression, we use OUTPUT as
' the input and ORIGINAL as the output filenames.
'
' Runs of matching characters that are 4 or more in length are
' compressed into the following format:
'
'       (char) (RLE) (count)
'

CONST _
  FALSE = 0, TRUE = NOT FALSE, _  ' standard boolean constants
  FirstTime  = 0, _             ' don't consider previous input
  LoneChar   = 1, _             ' lastchar sent by itself
  SendNewC   = 2, _             ' run done, send new char next
  SendRLECnt = 3, _             ' send RLE count next
  Repeating  = 4, _
  RLE = &H90, _                 ' repeat sequence identifier
  MaxCount = 255                ' max repeat sequence length

DIM SHARED State AS INTEGER     ' current RLE State
DIM SHARED PrevC AS STRING*1    ' value returned on last call
DIM SHARED C     AS STRING*1    ' current processing byte
DIM SHARED Count AS INTEGER     ' repeat sequence counter
DIM SHARED SmallCount AS INTEGER ' counter for less than minimum
                                ' match length runs
DIM SHARED LastPass AS INTEGER   

DECLARE SUB CompressInit (Inf AS STRING, OutF AS STRING)
DECLARE SUB CompressFile ()
DECLARE SUB UnCompressBytes ()
DECLARE SUB UncompressFile ()
DECLARE SUB CompressDone ()
DECLARE FUNCTION CompressByte$ ()

' the main program

DIM InputFile AS STRING
DIM OutputFile AS STRING

CLS
InputFile = "COMPRESS.BAS"
OutputFile = "OUTPUT"
PRINT "Compressing File: "; InputFile; " into "; OutputFile
CompressInit InputFile, OutputFile
CompressFile
CompressDone

InputFile = "OUTPUT"
OutputFile = "ORIGINAL"
PRINT "Decompressing File: "; InputFile; " into "; OutputFile
CompressInit InputFile, OutputFile
UnCompressFile
CompressDone
END

SUB CompressInit(Inf AS STRING, OutF AS STRING)
OPEN Inf FOR BINARY AS #1
OPEN OutF FOR BINARY AS #2
State = FirstTime
LastPass = FALSE
END SUB

FUNCTION CompressByte$
SELECT CASE State
  CASE FirstTime                ' no sequence in progress
    State = LoneChar
    SmallCount = 0
    GET #1,,PrevC
    CompressByte$ = PrevC
    EXIT FUNCTION

' the last input character was sent to the output file,
' now see if we can find a run of characters that match
' the last one output

  CASE LoneChar
    SELECT CASE ASC(PrevC)
      CASE RLE                  ' if we sent a real RLE
        State = FirstTime       ' then reset our state machine
        CompressByte$ = CHR$(0) ' and indicate the RLE was Real
        EXIT FUNCTION
      CASE ELSE                 ' else test for a repeat
        IF SmallCount = 0 THEN
          Count = 1
          GET #1,,C
          DO WHILE (C = PrevC) AND (Count < MaxCount) AND _
                (LOC(1) < LOF(1))
            Get #1,,C
            Count = Count + 1
          LOOP
          IF (C = PrevC) AND _
             (LOC(1) = LOF(1)) THEN
            LastPass = TRUE
            Count = Count + 1
          END IF
        ELSE
          SmallCount = SmallCount - 1  ' if smallcount=0 then
          IF SmallCount = 0 THEN       ' return the first non-
            PrevC = C                  ' match char, which we
            CompressByte$ = C          ' have already read from
            EXIT FUNCTION              ' Inf
          ELSE                         ' Added to published code***
            CompressByte$ = PrevC      ' chars < min match len
            EXIT FUNCTION
          END IF
        END IF
        SmallCount = 0                ' reset small run counter
        SELECT CASE Count
          CASE 1                      ' no match, or run of two,
            PrevC = C                 ' ...no compress
            CompressByte$ = C
            EXIT FUNCTION
          CASE 2                      ' a run of 3, no compress
            SmallCount = 1            ' send 1 PrevC next time
            CompressByte$ = PrevC
            EXIT FUNCTION
          CASE ELSE                   ' have a run of 4 or more
            State = SendRLECnt        ' send repeat count next
            CompressByte$ = CHR$(RLE) ' send RLE identifier
            EXIT FUNCTION
        END SELECT
      END SELECT

  CASE SendRLECnt                ' sent RLE, now send count and
    State = SendNewC             '  reset state machine to send
    CompressByte$ = CHR$(Count)  '  the first non-match char
    EXIT FUNCTION                '  which we have already read

  CASE SendNewC                  ' now send the next character
    State = LoneChar             '   that we have already read
    PrevC = C                    '   when searching for a run
    CompressByte$ = C            '   and make it the previous
    EXIT FUNCTION                '   char
END SELECT
END FUNCTION

SUB CompressFile
DIM CompChr AS STRING*1
DO WHILE LOC(1) < LOF(1)
  CompChr = CompressByte$             ' get next byte to output
  PUT #2,,CompChr
LOOP
IF LastPass THEN                      ' finish comressing after
  CompChr = CompressByte$             '  reaching the EOF
  PUT #2,,CompChr
END IF
END SUB

SUB UnCompressBytes
SELECT CASE State
  CASE FirstTime
    IF C = CHR$(RLE) THEN       ' when you see RLE for 1st time
      State = Repeating         '   just set the state
    ELSE                        ' if not RLE then output the
      PrevC = C                 '   character you just read
      PUT #2,,C
    END IF
  CASE Repeating
    IF ASC(C) > 0 THEN
      DO                        ' write out rest of characters
        C = CHR$(ASC(C) - 1)
        PUT #2,,PrevC
      LOOP UNTIL ASC(C) = 1
    ELSE                        ' output RLE as a real data byte
      C = CHR$(RLE)
      PUT #2,,C
    END IF
    State = FirstTime
END SELECT
END SUB

SUB UncompressFile
DO WHILE LOC(1) < LOF(1)
  GET #1,,C                     ' get next byte and go
  UnCompressBytes               '   uncompress it
LOOP
END SUB

SUB CompressDone
CLOSE #1
CLOSE #2
END SUB
