'======================================================================
' Program: LZW.BAS                                                    =
' Author: Gary Conway and Blake Ragsdell                              =
'                                                                     =
' This program compresses/uncompresses files using LZW compression    =
'======================================================================

DEFINT A-Z

CONST FALSE = 0, TRUE = NOT FALSE
CONST MaxTableEntries = 3000
CONST Prefix = 0, Suffix = 1, Link = 2

DECLARE SUB OutputUnCompValue (I%)
DECLARE SUB OutputCompValue (I%)
DECLARE SUB CompressFile ()
DECLARE SUB Done ()
DECLARE SUB Init (Inf AS STRING, Outf AS STRING)
DECLARE SUB InsertEntry ()
DECLARE SUB StoreTable (FileTbl$)
DECLARE SUB UnCompressFile ()
DECLARE FUNCTION InputUnCompValue% ()
DECLARE FUNCTION InputCompValue% ()
DECLARE FUNCTION ManageTbl% ()
DECLARE FUNCTION ExpandValue% (InputValue%, OutputStatus%)

'======================================================================
' If you would like to create a file with a copy of the string        =
' table LZWUnit used to compress the input file just define the       =
' Debugging value in the next line                                    =
'======================================================================

CONST Debugging = FALSE

OPTION BASE 0

DIM InputFile AS STRING
DIM SHARED CodeTbl(MaxTableEntries, 2) AS INTEGER
DIM SHARED HitTbl(MaxTableEntries) AS INTEGER
DIM SHARED PrefixCandidate AS INTEGER
DIM SHARED SuffixCandidate AS INTEGER
DIM SHARED TableFull AS INTEGER
DIM SHARED TableTop AS INTEGER
DIM SHARED InputFileNbr AS INTEGER
DIM SHARED OutputFileNbr AS INTEGER

CLEAR,,4096                 ' Make some extra stack space available

  CLS
  InputFile = COMMAND$

  IF LEN(InputFile) = 0 THEN
    PRINT "Please include a filename on the command line."
    STOP
  END IF

  PRINT "Compressing "; InputFile; " into OUTPUT.BAS"
  Init InputFile, "OUTPUT.BAS"
  CompressFile
  Done
  IF Debugging THEN StoreTable "CompTbl.BAS"

  CLS
  PRINT "Uncompressing OUTPUT.BAS into RESTORED.BAS"
  Init "OUTPUT.BAS", "RESTORED.BAS"
  UnCompressFile
  Done
  IF Debugging THEN StoreTable "CompTblU.BAS"

END

SUB CompressFile '=====================================================
'  CompressFile manages all the compression tasks                     =
'======================================================================

  Cur& = 0                   'Counter for screen

  PrefixCandidate = InputUnCompValue

LOOP1:

  Cur& = Cur& + 1                      'Manage counter display
  IF (Cur& AND 127) = 127 THEN
    LOCATE 10, 10
    PRINT USING "###,###"; Cur&;
  END IF

  SuffixCandidate = InputUnCompValue

  IF EOF(1) THEN
    OutputCompValue PrefixCandidate
    EXIT SUB
  END IF

  FoundCode = ManageTbl                     'Search table for P-S pair

  IF FoundCode THEN                         'If pair found
    PrefixCandidate = FoundCode             '  go look for next pair
  ELSE
    OutputCompValue (PrefixCandidate)       'otherwise, output Prefix
    PrefixCandidate = SuffixCandidate       '  and reset for next pair
  END IF

  GOTO LOOP1

END SUB

SUB Done 'Done closes the files========================================

  CLOSE 1, 2

END SUB

FUNCTION ExpandValue% (InputValue%, OutputStatus%) '===================
'  ExpandValue expands compression codes. Note, if the prefix value   =
'    retrieved in KPrefix is another compression code, ExpandValue    =
'    will recursively call itself until KPrefix is an extended ASCII  =
'    character.                                                       =
'                                                                     =
'  Input:                                                             =
'    InputValue is the value to expand                                =
'    OutputValue turns on/off writing of expanded characters to       =
'      file so you can retrieve (without writing) the first ASCIi     =
'      character at the head of the compressed character chain. This  =
'      becomes necessary when you must fill in the Suffix value in    =
'      string table for adjacent Prefix pointers.                     =
'  Output:                                                            =
'    Returns the character at the head of compressed byte chain when  =
'      you pass it a compressed byte. If you pass it an ASCII         =
'      character, it returns that character. This made coding simpler =
'======================================================================

IF InputValue > 255 THEN                          'Is compressed value?
  KPrefix = CodeTbl(InputValue, Prefix)           'Yes, get table entry
  KSuffix = CodeTbl(InputValue, Suffix)
  IF KPrefix > 255 THEN                               'If prefix is a
    KReturned = ExpandValue(KPrefix, OutputStatus)    ' compressed char
  ELSE      'set head value and output uncompressed bytes if output set
    KReturned = KPrefix
    IF OutputStatus THEN OutputUnCompValue (KPrefix)
  END IF
  IF OutputStatus THEN OutputUnCompValue (KSuffix)
ELSE
  KReturned = InputValue    'Return ASCII value if passed ASCII value
END IF
ExpandValue% = KReturned
END FUNCTION

SUB Init (Inf AS STRING, Outf AS STRING) '=============================
'  Initializes the string code table with the atomic values           =
'    and the table management values TableFull, TableTop, HitTbl()    =
'  HitTbl is an array that contains the head pointer for linked       =
'    lists (yes multiple lists) of compression code to facilitate     =
'    faster lookup of PrefixCandidate-SuffixCandidate pairs.          =
'  If HitTbl(Prefix value) = 0 then no P-S entries with the Prefix    =
'    value have been added to the string table. If                    =
'    HitTbl(Prefix value) <> 0, it contains the entry number of the   =
'    first element in the string table with that Prefix value         =
'  The CodeTbl(X,Link) element will contain a 0 if the string table   =
'    does not have any other entries that start with the prefix in    =
'    CodeTbl(X,Prefix), otherwise CodeTbl(X,Link) points to the next  =
'    entry with a matching PrefixCandidate value                      =
'======================================================================

  TableFull = FALSE          'Track Table status
  TableTop = 255             'Last entry number in table

  FOR I = 0 TO MaxTableEntries 'init string table for atomic characters
    HitTbl(I) = 0
    CodeTbl(I, Link) = 0
    IF I > 255 THEN
      CodeTbl(I, Prefix) = 0
      CodeTbl(I, Suffix) = 0
    ELSE
      CodeTbl(I, Prefix) = -1
      CodeTbl(I, Suffix) = I
    END IF
  NEXT

  InputFileNbr = FREEFILE
  OPEN Inf FOR BINARY AS #InputFileNbr
  OutputFileNbr = FREEFILE
  OPEN Outf FOR BINARY AS #OutputFileNbr

END SUB

FUNCTION InputCompValue% '=============================================
  GET InputFileNbr, , I%
  InputCompValue% = I%
END FUNCTION

FUNCTION InputUnCompValue% '===========================================
  C$ = " "
  GET InputFileNbr, , C$
  InputUnCompValue% = ASC(C$)
END FUNCTION

SUB InsertEntry '======================================================
' Insert PrefixCandidate-SuffixCandidate into the next available      =
'   entry in the table                                                =
'======================================================================

  CodeTbl(TableTop, Prefix) = PrefixCandidate
  CodeTbl(TableTop, Suffix) = SuffixCandidate
  IF TableTop = MaxTableEntries THEN TableFull = TRUE
END SUB

           '===========================================================
FUNCTION ManageTbl                                                   '=
'  ManageTbl searches the Table for PrefixCandidate-SuffixCandidate   =
'    pairs of characters/codes. If the pair is not in the string      =
'    table, it adds them and updates the linked list (See INIT)       =
'    If the pair is found, it returns the entry number for the pair.  =
'======================================================================

  Found = FALSE: EndOfLinks = FALSE    'Initialize table mngment values

  IF HitTbl(PrefixCandidate) <> 0 THEN 'Entries exist for Prefix
    CurPtr = HitTbl(PrefixCandidate) 'Trace list starting at head
    DO
      IF CodeTbl(CurPtr, Prefix) = PrefixCandidate _
       AND CodeTbl(CurPtr, Suffix) = SuffixCandidate THEN
        Found = TRUE
      ELSE
        IF CodeTbl(CurPtr, Link) <> 0 THEN
          CurPtr = CodeTbl(CurPtr, Link)
        ELSE
          EndOfLinks = TRUE
        END IF
      END IF
    LOOP UNTIL Found OR EndOfLinks
  END IF
  
  IF Found THEN                                'If pair found
    ManageTbl = CurPtr                         '  return element #
  ELSE                                         'otherwise, add to table
    IF NOT TableFull THEN
      TableTop = TableTop + 1                  'Set for new link #
      InsertEntry
      IF HitTbl(PrefixCandidate) = 0 THEN    'Adjust links
        HitTbl(PrefixCandidate) = TableTop
      ELSE
        CodeTbl(CurPtr, Link) = TableTop
      END IF
    END IF
    ManageTbl = FALSE
  END IF
END FUNCTION

SUB OutputCompValue (I%) '=============================================
  PUT OutputFileNbr, , I%
END SUB

SUB OutputUnCompValue (I%) '===========================================
  C$ = CHR$(I%)
  PUT OutputFileNbr, , C$
END SUB

SUB StoreTable (FileTbl$) '============================================
' OutputTable dumps a formatted list of the string table into the FILE=
'   specified in F.                                                   =
'======================================================================

  FileNbr = FREEFILE
  OPEN "O", FileNbr, FileTbl$

  FOR I = 256 TO TableTop

    PRINT #FileNbr, USING "####  "; I;
    PRINT #FileNbr, USING "####  "; CodeTbl(I, Prefix);
    PRINT #FileNbr, USING "####  "; CodeTbl(I, Suffix);
    PRINT #FileNbr, USING "####  "; CodeTbl(I, Link);

    FOR j = Prefix TO Suffix
      IF CodeTbl(I, j) < 255 AND CodeTbl(I, j) >= 32 THEN
        PRINT #FileNbr, USING "! "; CHR$(CodeTbl(I, j));
      ELSE
        PRINT #FileNbr, "  ";
      END IF
    NEXT j

    PRINT #FileNbr, " "

  NEXT I

  CLOSE FileNbr

END SUB

SUB UnCompressFile '===================================================
' UnCompresFile manages all aspects of uncompressing files            =
'======================================================================

Ctr& = 0
PrefixCandidate = InputCompValue

WHILE NOT EOF(1)

  Ctr& = Ctr& + 1                           'Manage onscreen display
  IF (Ctr& AND 127) = 127 THEN
    LOCATE 10, 10
    PRINT USING "###,###   "; Ctr&;
  END IF

  'Output an ASCII character

  IF PrefixCandidate < 256 THEN OutputUnCompValue (PrefixCandidate)
  SuffixCandidate = InputCompValue
  IF SuffixCandidate > 255 THEN      'Compressed character?  
    SuffixCopy = SuffixCandidate     'Save just in case we expand it

    'Handle special case when you need to expand an entry that you
    ' have not yet added to table                                 

    IF TableTop + 1 = SuffixCandidate THEN
      SuffixCandidate = ExpandValue(PrefixCandidate, FALSE)
      Found = ManageTbl
      SuffixCandidate = SuffixCopy
      Dummy = ExpandValue(SuffixCandidate, TRUE)
    ELSE
      SuffixCandidate = ExpandValue(SuffixCandidate, TRUE) 'Normal
      Found = ManageTbl                                    'expand
      SuffixCandidate = SuffixCopy
    END IF
  ELSE
    Found = ManageTbl
  END IF

  PrefixCandidate = SuffixCandidate

  IF EOF(1) THEN EXIT SUB              ' Leave if EOF

WEND

END SUB


